<?php

declare(strict_types=1);

namespace Drupal\Tests\panther\FunctionalJavascript;

use Drupal\user\UserInterface;
use Facebook\WebDriver\Exception\NoSuchElementException;
use Facebook\WebDriver\Exception\StaleElementReferenceException;
use Facebook\WebDriver\WebDriver;
use Facebook\WebDriver\WebDriverBy;
use Facebook\WebDriver\WebDriverElement;
use Facebook\WebDriver\WebDriverExpectedCondition;
use Facebook\WebDriver\WebDriverKeys;

/**
 * Provides Drupal-related methods.
 */
trait DrupalTrait {

  /**
   * Logs in as a user.
   *
   * @param string $user
   *   The user's name.
   * @param string $password
   *   The password.
   * @param string $login_path
   *  The path where the login form is located.
   */
  public function loginAs(string $user, string $password, string $login_path = '/user/login'): void {
    $this->goToPage($login_path);
    $this->maximizeWindow();
    self::assertPageContains('Log in');

    $this->submitForm('Log in', [
      'name' => $user,
      'pass' => $password,
    ]);
    $this->takeScreenshots('admin', 'admin');
  }

  /**
   * Select an operation on a table row with a given text.
   *
   * @param string $text
   * @param int $column
   * @param string $operation
   */
  public function selectOperationOnTableRowWithText(
    string $text,
    int $column,
    string $operation,
  ): void {
    $value = sprintf(
      '//table//tr[td[%d][contains(., \'%s\')]]',
      $column,
      $text
    );

    $this->scrollDownUntilVisible(WebDriverBy::xpath($value));

    $crawler = $this
      ->getClient()
      ->getCrawler();

    $row = $crawler->filterXPath($value);
    $row->filter('.dropbutton__toggle')->click();

    try {
      $this->getClient()->waitFor('.dropbutton__items');
      $row->selectLink($operation)->click();

      // Reload the page to reset the crawler state.
      $this->getClient()->reload();
    }
    catch (\Exception $e) {
      $this->fail($e->getMessage());
    }
  }

  /**
   * Select an option from a select element whose label is given.
   *
   * @param string $label
   *   The label of the select element.
   * @param string $value
   *   The value of the option to select.
   */
  public function selectOption(string $label, string $value): void {
    $crawler = $this->getClient()->getCrawler();

    $xpath = sprintf(
      ".//label[text()[contains(.,'%s')]][1]/following-sibling::select[1]//option[@value='%s']",
      $label,
      $value
    );
    $option = $crawler->filterXPath($xpath);
    $option->click();
  }

  /**
   * Scroll down until the given element is visible.
   *
   * @param \Facebook\WebDriver\WebDriverBy $by
   *   The WebDriverBy object.
   */
  public function scrollDownUntilVisible(WebDriverBy $by): void {
    try {
      $element = $this->getClient()->findElement($by);
      while (!$element->isDisplayed()) {
        $this->getClient()->getKeyboard()->pressKey(WebDriverKeys::PAGE_DOWN);
      }
    }
    catch (\Exception $e) {
      $this->fail($e->getMessage());
    }
  }

  /**
   * Scroll down until the given element is visible.
   *
   * @param string $selector
   *   The CSS selector.
   */
  public function scrollDownUntilVisibleBySelector(string $selector): void {
    $this->scrollDownUntilVisible(WebDriverBy::cssSelector($selector));
  }

  /**
   *
   */
  public function clickSubmitButton(): void {
    $this->waitForElementToBeClickable(".//input[@id='edit-submit']")->click();
  }

  /**
   * Create a user.
   *
   * @param \stdClass $user
   */
  public function userCreate(
    \stdClass $user,
  ): void {
    $this->driver->userCreate($user);
    $this->createdEntityManager->addUser($user);
  }

  /**
   * Create a role.
   *
   * @param array $permissions
   *
   * @return string
   */
  public function roleCreate(
    array $permissions,
  ): string {
    $role = $this->driver->roleCreate($permissions);
    $this->createdEntityManager->addRole($role);

    return $role;
  }

  /**
   * Delete a user by email.
   *
   * @param string $email
   *   The user's email.
   */
  public function deleteUserByEmail(string $email): void {
    try {
      /** @var \Drupal\user\UserInterface[] $users */
      $users = \Drupal::entityTypeManager()
        ->getStorage('user')
        ->loadByProperties(['mail' => $email]);

      if (\count($users) > 0) {
        $user = reset($users);
        $this->driver->userDelete((object) ['uid' => $user->id()]);
        $this->driver->processBatch();
      }
    }
    catch (\Exception $e) {
      $this->fail($e->getMessage());
    }
  }

  /**
   * Create a user with a given role and fields and then log in as that user.
   *
   * @param string $role
   *   The role.
   * @param array $fields
   *   The user fields.
   * @param string $login_path
   *   The path where the login form is located.
   */
  public function loginAsUserByRole(string $role, array $fields = [], string $login_path = '/user/login'): void {
    $name = $this->driver->getRandom()->name(8);
    $password = $this->driver->getRandom()->name(16);

    $user = (object) [
      'name' => $name,
      'pass' => $password,
      'role' => $role,
    ];
    $user->mail = "{$user->name}@example.com";

    foreach ($fields as $field => $value) {
      $user->{$field} = $value;
    }

    $this->userCreate($user);

    $roles = explode(',', $role);
    $roles = array_map('trim', $roles);
    foreach ($roles as $role) {
      if (!in_array(strtolower($role), [
        'authenticated',
        'authenticated user',
      ])) {
        // Only add roles other than 'authenticated user'.
        $this->driver->userAddRole($user, $role);
      }
    }

    $this->loginAs($name, $password, $login_path);
  }

  public function loginAsUserWithPermissions(array $permissions, array $fields = [], string $login_path = '/user/login'): void {
    $role = $this->roleCreate($permissions);

    $this->loginAsUserByRole($role, $fields, $login_path);
  }

  /**
   *
   *
   * @param \Drupal\user\UserInterface $user
   * @param string $password
   */
  public function setUserPassword(UserInterface $user, string $password): void {
    try {
      $user->setPassword($password)->save();
    }
    catch (\Exception $e) {
      $this->fail($e->getMessage());
    }
  }

  /**
   * @param int $timeout
   * @param string $condition
   *
   * @return bool
   */
  public function waitOnJavascript(int $timeout, string $condition): bool {
    $script = 'return (' . rtrim($condition, " \t\n\r;") . ');';

    try {
      $start = microtime(TRUE);
      $end = $start + $timeout / 1000.0;
      do {
        $result = $this->getClient()->executeScript($script);
        if ($result) {
          break;
        }
        usleep(10000);
      } while (microtime(TRUE) < $end);

      return (bool) $result;
    }
    catch (\Exception $e) {
      $this->fail($e->getMessage());
    }
  }

  /**
   * @param int $timeout
   *
   * @return bool
   */
  public function waitForAjaxToFinish(int $timeout = 1000): bool {
    $condition = <<<JS
    (function() {
      function isAjaxing(instance) {
        return instance && instance.ajaxing === true;
      }
      var d7_not_ajaxing = true;
      if (typeof Drupal !== 'undefined' && typeof Drupal.ajax !== 'undefined' && typeof Drupal.ajax.instances === 'undefined') {
        for(var i in Drupal.ajax) { if (isAjaxing(Drupal.ajax[i])) { d7_not_ajaxing = false; } }
      }
      var d8_not_ajaxing = (typeof Drupal === 'undefined' || typeof Drupal.ajax === 'undefined' || typeof Drupal.ajax.instances === 'undefined' || !Drupal.ajax.instances.some(isAjaxing))
      return (
        // Assert no AJAX request is running (via jQuery or Drupal) and no
        // animation is running.
        (typeof jQuery === 'undefined' || jQuery.hasOwnProperty('active') === false || (jQuery.active === 0 && jQuery(':animated').length === 0)) &&
        d7_not_ajaxing && d8_not_ajaxing
      );
    }());
JS;

    return $this->waitOnJavascript($timeout, $condition);
  }

  /**
   * Wait for a DOM element to appear.
   *
   * @param string $selector
   *   The CSS selector.
   *
   * @return \Facebook\WebDriver\WebDriverElement|null
   *
   * @throws \Facebook\WebDriver\Exception\NoSuchElementException
   * @throws \Facebook\WebDriver\Exception\TimeoutException
   */
  public function waitForDomElement(string $selector): ?WebDriverElement {
    $this->getClient()->wait()->until(
      static function (WebDriver $driver) use ($selector) {
        try {
          return $driver->findElement(WebDriverBy::cssSelector($selector));
        }
        catch (StaleElementReferenceException | NoSuchElementException $e) {
          return NULL;
        }
      }
    );

    return NULL;
  }

  /**
   * Expand the toolbar.
   */
  public function expandToolbar(): void {
    $this->getClient()->getCrawler()->filter('.toolbar-menu__trigger')->click();
  }

  /**
   * Select a checkbox for a given data-drupal-selector.
   *
   * @param string $fieldset_label
   *   The fieldset's label which contains the checkbox.
   * @param string $checkbox_label
   *   The checkbox's label.
   */
  public function selectCheckbox(string $fieldset_label, string $checkbox_label): void {
    $selector = sprintf(
      ".//span[@class[contains(., 'fieldset__label')] and text()[contains(., '%s')]]/ancestor::fieldset//label[text()[contains(., '%s')]]/parent::div//input[@class[contains(., 'form-checkbox')]]",
      $fieldset_label,
      $checkbox_label
    );

    $this->waitForElementToBeClickable($selector)->click();
  }

  /**
   * Click on a field settings cog button.
   *
   * @param string $label
   *   The field's label.
   */
  public function clickFieldCogButton(string $label): void {
    $selector = sprintf(
      ".//div[@class='tabledrag-cell-content__item' and text()[contains(., '%s')]]/ancestor::tr//input[@class[contains(., 'field-plugin-settings-edit')]]",
      $label
    );

    $this->waitForElementToBeClickable($selector)->click();
  }

  /**
   * Wait for an element to be clickable.
   *
   * Sometimes the element you're trying to click is being covered by another
   * element, specifically a sticky region div. This method will scroll the
   * element into view and wait for it to be clickable.
   *
   * @param string $selector
   *   The xPath selector.
   *
   * @return \Facebook\WebDriver\WebDriverElement
   *   The element.
   */
  public function waitForElementToBeClickable(string $selector): WebDriverElement {
    $web_driver = WebDriverBy::xpath($selector);

    try {
      $element = $this->getClient()->findElement($web_driver);
      $this->getClient()->executeScript('arguments[0].scrollIntoView(true);', [$element]);
      $this->getClient()->wait()->until(
        WebDriverExpectedCondition::elementToBeClickable($web_driver)
      );
    }
    catch (\Exception $e) {
      $this->fail($e->getMessage());
    }

    return $element;
  }
}
